#!/usr/bin/env python

import Image, ImageTk, Tkinter, sys, time, re, tkFileDialog
from math import *
from optparse import OptionParser

class gaussBoxTool:
    def __init__(self, posX, posY, color, target):
        self.boxIdR = 0
        self.boxIdL = 0
        self.boxX = posX
        self.boxY = posY
        self.target = target
        self.color = color

    def draw(self, posX):
        width = fabs(self.boxX - posX)
        if (self.boxIdR != 0): self.target.delete(self.boxIdR)
        if (self.boxIdL != 0): self.target.delete(self.boxIdL)
        self.boxIdR = self.target.create_rectangle(self.boxX, self.boxY, self.boxX+width, self.boxY+palh-1, outline=self.color)
        self.boxIdL = self.target.create_rectangle(self.boxX-width, self.boxY, self.boxX, self.boxY+palh-1, outline=self.color)

    def hide(self):
        self.target.delete(self.boxIdR)
        self.target.delete(self.boxIdL)

class Application:
    def gauss(self, x, mu, sigma, amp=255):
        return int(amp*exp(-(x-mu)*(x-mu)/(sigma*sigma)))

    def getColor(self, i):
        q = float(i)/float(palw)
        index = {'red':0, 'green':1, 'blue':2}
        rgb = [0,0,0]
        for curve in self.curves:
            rgb[index[curve[0]]] += self.gauss(q, curve[1], curve[2], curve[3])
        return (rgb[0],rgb[1],rgb[2])

    def setColors(self):
        self.color = []
        if (len(self.curves) == 0):
            print 'blank palette'
            self.color = [(0,0,0)]*palw
        else:
            print 'setting up palette with', palw, 'colors'
            for i in range(palw):
                self.color.append(self.getColor(i))        

    def draw(self):
        for x in range(palw):
            color = self.color[x]
            self.image.paste(color, (x, 0, x+1, palh))

        for i in range(3):
            for x in range(palw):
                index = [0, 0, 0]
                index[i] = 1
                color = (self.color[x][i]*index[0], self.color[x][i]*index[1], self.color[x][i]*index[2])
                self.image.paste(color, (x, (i+1)*(palh+pals), x+1, (i+2)*palh+(i+1)*pals))
        self.imgPal = ImageTk.PhotoImage(self.image)
        self.canPal.create_image(palw/2, palth/2, image=self.imgPal)
        self.canPal.update_idletasks()

    def removeCurve(self, band, pos):
        bestDist = pals
        bestIndex = -1
        for i in range(len(self.curves)):
            iBand = self.curves[i][0]
            iMu   = self.curves[i][1]
            iSig  = self.curves[i][2]
            if ((iBand == band) and (fabs(pos - iMu) < iSig)):
                if (fabs(pos - iMu) < bestDist):
                    bestDist = fabs(pos - iMu)
                    bestIndex = i
        if (bestIndex != -1):            
            popped = self.curves.pop(bestIndex)
            self.infoMsg('remove gaussian %s %1.3f %1.3f %i' % popped)
            self.reset()
        else:
            print 'no match!'

    def load(self):
        self.infoMsg('open palette')
        fileFormat = [('Palette file (ascii)','*.pal')]
        fileH = tkFileDialog.askopenfile(parent=self.parent,
                                            filetypes=fileFormat,
                                            mode='r',
                                            title="Open palette")
        if fileH != None:
            self.curves = []
            for line in fileH.read().splitlines():
                print '\t'+line
                fields = line.split()
                amp = 255
                if (len(fields)==4): amp = int(fields[3])
                self.curves.append((fields[0], float(fields[1]), float(fields[2]), amp))
            fileH.close()
            self.reset()
        

    def reset(self):
        self.setColors()
        self.draw()

    def blank(self):
        self.infoMsg('reset')
        self.curves = []
        self.reset()

    def undo(self):
        if (len(self.curves)==0):
            self.infoMsg('nothing to undo!')
            return
        self.infoMsg('undo!')
        self.curves.pop()
        self.reset()

    def __init__(self, parent):
        self.parent = parent
        self.image = Image.new("RGB", (palw, palth), (127,127,127))
        self.imgPal = ImageTk.PhotoImage(self.image)
        self.createWidgets()
        self.blank()
        self.infoMsg('ready')
        self.busy = False
        self.selX = 0.0
        self.selBand = ''
        self.amp = 255
        
    def infoMsg(self, msg):
        self.labInfo.configure(text=msg)
        print msg

    def save(self):
        fileFormat = [('Palette file (ascii)','*.pal'), ('jpeg file', '*.jpg')]
        fileName = tkFileDialog.asksaveasfilename(parent=self.parent,
                                                  filetypes=fileFormat,
                                                  title="Save as")
        if len(fileName) > 0:
            fileH = open(fileName, 'w')
            if (re.search('.jpg', fileName)):
                self.image.crop((0,0,palw,palh)).save(fileH, 'JPEG')
            if (re.search('.pal', fileName)):
                fileBody = ''
                for curve in self.curves:
                    fileBody += '%s\t%f  %f  %i\n' % curve
                fileH.write(fileBody)
            fileH.close()
            self.infoMsg("saved to %s" % fileName)

    def evtMouseDown(self, event):
        colDict = {0:'red', 1:'green', 2:'blue'}
        band=-1
        for i in range(3):
            limitDown = (palh+pals)*(i+1)
            limitUp = palh*(i+2)+pals*(i+1)
            if (event.y > limitDown and event.y < limitUp):
                band = i
        if (band != -1):
            bandName = colDict[band]
            gCenter = float(event.x)/float(palw)
            self.busy = True
            self.selX = gCenter
            self.selBand = bandName
            self.gaussBox = gaussBoxTool(event.x, (palh+pals)*(band+1), bandName, self.canPal)

    def evtMouseMove(self, event):
        colDict = {0:'all', 1:'red', 2:'green', 3:'blue'}
        if (self.busy):
            curX = float(event.x)/float(palw)
            curW = fabs(self.selX - curX)
            self.labPos.configure(text='gaussian width %f' % curW)
            self.gaussBox.draw(event.x)
        else:
            for i in range(4):
                limitDown = (palh+pals)*i
                limitUp = palh*(i+1)+pals*i
                if (event.y > limitDown and event.y < limitUp):
                    colName = colDict[i]
                    self.labPos.configure(text='%s %s' % (colName, float(event.x)/float(palw)))
                    break
                else:
                    self.labPos.configure(text='')
                
    def evtMouseUp(self, event):
        if (self.busy == False): return
        curX = float(event.x)/float(palw)
        curW = fabs(self.selX - curX)
        self.gaussBox.hide()
        del self.gaussBox
        if (curW == 0):
            self.busy = False
            return
        self.curves.append((self.selBand, self.selX, curW, self.amp))
        self.infoMsg('add gaussian %s %1.3f %1.3f %i' % self.curves[-1])
        self.busy = False
        self.reset()
        
    def evtMouseRight(self, event):
        colDict = {0:'red', 1:'green', 2:'blue'}
        band=-1
        for i in range(3):
            limitDown = (palh+pals)*(i+1)
            limitUp = palh*(i+2)+pals*(i+1)
            if (event.y > limitDown and event.y < limitUp):
                band = i
        if (band != -1):
            bandName = colDict[band]
            curX = float(event.x)/float(palw)
            self.removeCurve(bandName, curX)

    def setAmp(self, event):
        self.amp = self.sclAmp.get()

    def createWidgets(self):
        
        self.frmTop = Tkinter.Frame(self.parent)
        self.frmTop.pack(anchor="w")
        self.frmPal = Tkinter.Frame(self.parent)
        self.frmPal.pack(anchor="w", in_=self.frmTop, side="left")

        self.canPal = Tkinter.Canvas(self.parent, width=palw, height=palth)        
        self.canPal.pack(in_=self.frmPal)
        self.canPal.bind("<Motion>", self.evtMouseMove)
        self.canPal.bind("<Button-1>", self.evtMouseDown)
        self.canPal.bind("<ButtonRelease-1>", self.evtMouseUp)
        self.canPal.bind("<ButtonRelease-3>", self.evtMouseRight)

        self.sclAmp = Tkinter.Scale(self.parent, to=255, orient='horizontal', command=self.setAmp)
        self.sclAmp.pack(in_=self.frmPal, fill='x')
        self.sclAmp.set(255)
        
        self.frm1 = Tkinter.Frame(self.parent)
        self.frm1.pack(anchor="w", fill='x')

        self.frm4 = Tkinter.Frame(self.parent)
        self.frm4.pack(anchor="n", in_=self.frmTop, side="left")

        self.labInfo = Tkinter.Label(self.parent, text="Initializing...")
        self.labInfo.pack(side="right", in_=self.frm1, fill='x')

        self.labPos = Tkinter.Label(self.parent, text="")
        self.labPos.pack(side="left", in_=self.frm1, fill='x')

        self.btnReset = Tkinter.Button(self.parent, text="Reset", command=self.blank)
        self.btnReset.pack(in_=self.frm4, fill='x')

        self.btnUndo = Tkinter.Button(self.parent, text="Undo", command=self.undo)
        self.btnUndo.pack(in_=self.frm4, fill='x')

        self.btnSave = Tkinter.Button(self.parent, text="Save", command=self.save)
        self.btnSave.pack(in_=self.frm4, fill='x')

        self.btnOpen = Tkinter.Button(self.parent, text="Open", command=self.load)
        self.btnOpen.pack(in_=self.frm4, fill='x')

        self.btnQuit = Tkinter.Button(self.parent, text="Exit", command=self.parent.quit)
        self.btnQuit.pack(in_=self.frm4, fill='x')


# options
parser = OptionParser()
parser.add_option("-s", "--size", type="int", dest="palw",
                  help="size of image in pixels", default=1000)
(options, args) = parser.parse_args()


palw = options.palw
palh = 80
pals = 8
palth = 4*palh+3*pals

App = Tkinter.Frame()
App.pack()
App.master.title("Palette Tool")
im = Application(App)

App.mainloop()

